import collections
import functools
import os
import re
import sys

try:
    import xml.etree.cElementTree as ElementTree
except ImportError:
    import xml.etree.ElementTree as ElementTree


def walk(root, xml_name_regex="[.]xml$"):
    """
    :param root: directory containing Anafora XML directories
    :param str xml_name_regex: regular expression identifying .xml files to include
    :return iterator: an iterator of (sub-dir, text-file-name, xml-file-names) where sub-dir is the path to the Anafora
        directory relative to root, text-file-name is the name of the Anafora text file, and xml-file-names is a list
        of names of Anafora XML files
    """
    for dir_path, dir_names, file_names in os.walk(root):
        if not dir_names:
            sub_dir = ''
            if dir_path.startswith(root):
                sub_dir = dir_path[len(root):]
                if sub_dir.startswith(os.path.sep):
                    sub_dir = sub_dir[len(os.path.sep):]
            xml_names = [file_name for file_name in file_names if re.search(xml_name_regex, file_name) is not None]
            if xml_names:
                text_name = os.path.basename(dir_path)
                yield sub_dir, text_name, xml_names


def walk_anafora_to_anafora(root, xml_name_regex="[.]xml$"):
    """
    :param str root: path of the root directory to be walked
    :param str xml_name_regex: regular expression identifying .xml files to include
    :return iterator: an iterator of (input-sub-dir, output-sub-dir, text-file-name, xml-file-names)
    """
    for sub_dir, text_name, xml_names in walk(root, xml_name_regex):
        yield sub_dir, sub_dir, text_name, xml_names


def walk_flat_to_anafora(text_dir):
    """
    :param str text_dir: path to a directory of text files (and no subdirectories)
    :return iterator: an iterator of (input-sub-dir, output-sub-dir, text-file-name, xml-file-names)
    """
    for file_name in os.listdir(text_dir):
        yield '', file_name, file_name, []


class _XMLWrapper(object):
    def __init__(self, xml):
        """
        :param xml.etree.ElementTree.Element xml: the XML element to be wrapped in an object
        """
        self.xml = xml

    def __repr__(self):
        if self.xml is not None:
            result = ElementTree.tostring(self.xml)
            if sys.version_info.major >= 3:
                result = result.decode('utf-8')
        else:
            result = '{0}()'.format(self.__class__.__name__)
        return result


class AnaforaData(_XMLWrapper):
    def __init__(self, xml=None):
        """
        :param xml.etree.ElementTree.Element xml: the <data> element
        """
        if xml is None:
            xml = ElementTree.Element("data")
        _XMLWrapper.__init__(self, xml)
        self.annotations = AnaforaAnnotations(self.xml.find("annotations"), self)

    @classmethod
    def from_file(cls, xml_path):
        try:
            return cls(ElementTree.parse(xml_path).getroot())
        except ElementTree.ParseError as e:
            raise ValueError("invalid XML file {0}: {1}".format(xml_path, e))

    def indent(self, string="\t"):
        # http://effbot.org/zone/element-lib.htm#prettyprint
        def _indent(elem, level=0):
            i = "\n" + level * string
            if len(elem):
                if not elem.text or not elem.text.strip():
                    elem.text = i + string
                if not elem.tail or not elem.tail.strip():
                    elem.tail = i
                for elem in elem:
                    _indent(elem, level + 1)
                if not elem.tail or not elem.tail.strip():
                    elem.tail = i
            else:
                if level and (not elem.tail or not elem.tail.strip()):
                    elem.tail = i
        _indent(self.xml)

    def to_file(self, xml_path):
        ElementTree.ElementTree(self.xml).write(xml_path, encoding="UTF-8", xml_declaration=True, short_empty_elements=False)


class AnaforaAnnotations(_XMLWrapper):
    def __init__(self, xml, _data):
        _XMLWrapper.__init__(self, xml)
        self._data = _data
        self._id_to_annotation = collections.OrderedDict()
        if self.xml is not None:
            for annotation_elem in self.xml:
                if annotation_elem.tag == "entity":
                    annotation = AnaforaEntity(annotation_elem, self)
                elif annotation_elem.tag == "relation":
                    annotation = AnaforaRelation(annotation_elem, self)
                else:
                    raise ValueError("invalid tag: {0}".format(annotation_elem.tag))
                if annotation.id in self._id_to_annotation:
                    raise ValueError("duplicate id: {0}".format(annotation.id))
                self._id_to_annotation[annotation.id] = annotation

    def __iter__(self):
        return iter(self._id_to_annotation.values())

    def append(self, annotation):
        """
        :param AnaforaAnnotation annotation: the annotation to add
        """
        if annotation.id is None:
            raise ValueError("no id defined for {0}".format(annotation))
        if annotation.id in self._id_to_annotation:
            raise ValueError("duplicate id: {0}".format(annotation.id))
        annotation._annotations = self
        if self.xml is None:
            self.xml = ElementTree.SubElement(self._data.xml, "annotations")
        self.xml.append(annotation.xml)
        self._id_to_annotation[annotation.id] = annotation

    def remove(self, annotation):
        """
        :param AnaforaAnnotation annotation: the annotation to remove
        """
        if annotation.id is None:
            raise ValueError("no id defined for {0}".format(annotation))
        self.xml.remove(annotation.xml)
        del self._id_to_annotation[annotation.id]

    def select_id(self, id):
        return self._id_to_annotation[id]

    def select_type(self, type_name):
        for annotation in self:
            if annotation.type == type_name:
                yield annotation

    def find_self_referential(self):
        for annotation in self:
            if annotation.is_self_referential():
                return annotation


@functools.total_ordering
class AnaforaAnnotation(_XMLWrapper):
    def __init__(self, xml, _annotations):
        """
        :param xml.etree.ElementTree.Element xml: xml definition of this annotation
        :param AnaforaAnnotations _annotations: the annotations collection containing this annotation
        """
        _XMLWrapper.__init__(self, xml)
        self._annotations = _annotations
        self.properties = AnaforaProperties(self.xml.find("properties"), self)

    def __eq__(self, other):
        return (
            isinstance(other, AnaforaAnnotation) and
            self.spans == other.spans and
            self.type == other.type and
            self.properties == other.properties)

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        result = hash(self.spans)
        result = 31 * result + hash(self.type)
        result = 31 * result + hash(self.properties)
        return result

    def __lt__(self, other):
        return self.spans < other.spans

    @property
    def id(self):
        return self.xml.findtext("id")

    @id.setter
    def id(self, value):
        id_elem = self.xml.find("id")
        if id_elem is None:
            id_elem = ElementTree.SubElement(self.xml, "id")
        id_elem.text = value

    @property
    def type(self):
        return self.xml.findtext("type")

    @type.setter
    def type(self, value):
        type_elem = self.xml.find("type")
        if type_elem is None:
            type_elem = ElementTree.SubElement(self.xml, "type")
        type_elem.text = value

    @property
    def parents_type(self):
        return self.xml.findtext("parentsType")

    @parents_type.setter
    def parents_type(self, value):
        parents_type_elem = self.xml.find("parentsType")
        if parents_type_elem is None:
            parents_type_elem = ElementTree.SubElement(self.xml, "parentsType")
        parents_type_elem.text = value

    @property
    def spans(self):
        raise NotImplementedError

    def is_self_referential(self, seen_ids=None):
        if seen_ids is None:
            seen_ids = {}
        seen_ids[id(self)] = self
        for name in self.properties:
            value = self.properties[name]
            if seen_ids is not None and id(value) in seen_ids:
                return True
            if isinstance(value, AnaforaAnnotation):
                if value.is_self_referential(dict(seen_ids)):
                    return True
        return False

class AnaforaProperties(_XMLWrapper):
    def __init__(self, xml, _annotation):
        """
        :param xml.etree.ElementTree.Element xml: a <properties> element
        :param AnaforaAnnotation _annotation: the annotation containing these properties
        """
        _XMLWrapper.__init__(self, xml)
        self._annotation = _annotation
        self._tag_to_property_xml = {}
        if self.xml is not None:
            for property_elem in self.xml:
                self._tag_to_property_xml[property_elem.tag] = property_elem

    def __eq__(self, other):
        if not isinstance(other, AnaforaProperties):
            return False
        for name in self:
            self_value = self[name]
            if name not in other._tag_to_property_xml:
                return False
            other_value = other[name]
            if self_value != other_value:
                return False
        for name in other:
            if name not in self._tag_to_property_xml:
                return False
        return True

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        result = 0
        for name in self:
            result += hash(name)
            result += hash(self[name])
        return result

    def __iter__(self):
        return iter(self._tag_to_property_xml)

    def __contains__(self, property_name):
        return property_name in self._tag_to_property_xml

    def __getitem__(self, property_name):
        value = self._tag_to_property_xml[property_name].text
        return self._annotation._annotations._id_to_annotation.get(value, value)

    def __setitem__(self, name, value):
        if isinstance(value, AnaforaAnnotation):
            if self._annotation is None or self._annotation._annotations is None:
                message = 'annotation must be in <annotations> before assigning annotation value to property "{0}":\n{1}'
                raise ValueError(message.format(name, self._annotation))
            if value != self._annotation._annotations._id_to_annotation.get(value.id):
                message = 'annotation must be in <annotations> before assigning it to property "{0}":\n{1}'
                raise ValueError(message.format(name, value))
        if self.xml is None:
            self.xml = ElementTree.SubElement(self._annotation.xml, "properties")
        property_elem = self.xml.find(name)
        if property_elem is None:
            old_tail = None
            if len(self.xml) > 0:
                old_tail = self.xml[-1].tail
                self.xml[-1].tail = self.xml.text
            property_elem = ElementTree.SubElement(self.xml, name)
            property_elem.tail = old_tail
            self._tag_to_property_xml[name] = property_elem
        if isinstance(value, AnaforaAnnotation):
            property_elem.text = value.id
        else:
            property_elem.text = value

    def __delitem__(self, name):
        if name not in self._tag_to_property_xml:
            raise ValueError('no such property {0!r}'.format(name))
        self.xml.remove(self._tag_to_property_xml.pop(name))
        if not self._tag_to_property_xml:
            self._annotation.xml.remove(self.xml)
            self.xml = None

    def items(self):
        return [(name, self[name]) for name in self]


class AnaforaEntity(AnaforaAnnotation):
    def __init__(self, xml=None, _annotations=None):
        if xml is None:
            xml = ElementTree.Element("entity")
        AnaforaAnnotation.__init__(self, xml, _annotations)

    @property
    def spans(self):
        spans_text = self.xml.findtext("span")
        if spans_text is None:
            return ()
        return tuple(tuple(int(offset) for offset in tuple(span_text.split(",")))
                     for span_text in spans_text.split(";"))

    @spans.setter
    def spans(self, spans):
        if not isinstance(spans, tuple) or not all(isinstance(span, tuple) and len(span) == 2 for span in spans):
            raise ValueError("spans must be a tuple of pairs")
        span_elem = self.xml.find("span")
        if span_elem is None:
            span_elem = ElementTree.SubElement(self.xml, "span")
        span_elem.text = ";".join("{0:d},{1:d}".format(*span) for span in spans)


class AnaforaRelation(AnaforaAnnotation):
    def __init__(self, xml=None, _annotations=None):
        if xml is None:
            xml = ElementTree.Element("relation")
        AnaforaAnnotation.__init__(self, xml, _annotations)

    @property
    def spans(self):
        return tuple(
            self.properties[name].spans
            for name in sorted(self.properties)
            if isinstance(self.properties[name], AnaforaAnnotation))



from __future__ import absolute_import

import argparse
import codecs
import collections
import json
import logging
import os

import regex

import anafora
import anafora.evaluate


class RegexAnnotator(object):

    _word_boundary_pattern = regex.compile(r'\b')
    _capturing_group_pattern = regex.compile(r'[^\\]\([^?]')

    @classmethod
    def from_file(cls, path_or_file):
        """
        :param str|file path_or_file: a string path or a file object containing a serialized RegexAnnotator
        """

        # if passed a path instead of a file, open the path as a file for reading and recursively call from_file
        try:
            path_or_file.readline
        except AttributeError:
            with codecs.open(path_or_file, 'r', 'utf-8') as output_file:
                return cls.from_file(output_file)
        else:

            # parse each line in the model file, which should be: <regex>\t<type>\t<attributes>
            regex_type_attributes_map = {}
            default_type_attributes_map = {}
            for line in path_or_file:
                items = line.rstrip().split("\t")
                if len(items) < 2 or len(items) > 3:
                    raise ValueError('expected {0!r}, found {1!r}'.format("<regex>\t<type>\t<attributes>", line))
                if len(items) == 2:
                    [expression, entity_type] = items
                    attributes = {}
                else:
                    [expression, entity_type, attributes_string] = items
                    attributes = json.loads(attributes_string)

                # make sure the <regex> is a valid regular expression and does not use any capturing groups
                try:
                    regex.compile(expression)
                except regex.error as e:
                    raise ValueError("{0} in {1!r}".format(e, expression))
                if cls._capturing_group_pattern.search(expression):
                    raise ValueError("capturing groups are not allowed: " + expression)

                # map the regular exrpression to the type and attributes
                if not expression:
                    default_type_attributes_map[entity_type] = attributes
                else:
                    regex_type_attributes_map[expression] = (entity_type, attributes)
            return cls(regex_type_attributes_map, default_type_attributes_map or None)

    @classmethod
    def train(cls, text_data_pairs, min_count=None):
        """
        Train a regex model from a set of annotated entities

        :param collections.Iterable text_data_pairs: an iterable of `(text, data)` pairs where `text` is a string and
            `data` is an anafora.AnaforaData object
        :param int min_count: the minimum number of matched entities required to keep a pattern
        """

        # prepare mappings from text to type and attributes
        ddict = collections.defaultdict
        text_type_map = ddict(lambda: collections.Counter())
        text_type_attrib_map = ddict(lambda: ddict(lambda: ddict(lambda: collections.Counter())))
        default_type_attrib_map = ddict(lambda: ddict(lambda: collections.Counter()))

        # save a regular expression for each entity
        for text, data in text_data_pairs:
            for annotation in data.annotations:
                if isinstance(annotation, anafora.AnaforaEntity):
                    annotation_text = ' '.join(text[begin:end] for begin, end in annotation.spans)
                    if annotation_text:

                        # convert text into a regular expression, with generic whitespace matchers
                        annotation_regex = r'\s+'.join(regex.escape(s) for s in annotation_text.split())

                        # prefix and suffix with word break matchers if appropriate
                        begin = min(begin for begin, end in annotation.spans)
                        prefix = r'\b' if cls._word_boundary_pattern.match(text, begin) is not None else ''
                        end = max(end for begin, end in annotation.spans)
                        suffix = r'\b' if cls._word_boundary_pattern.match(text, end) is not None else ''

                        # update counts of the regular expression and its corresponding type and attributes
                        annotation_regex = '{0}{1}{2}'.format(prefix, annotation_regex, suffix)
                        text_type_map[annotation_regex][annotation.type] += 1
                        for key, value in annotation.properties.items():
                            if not isinstance(value, anafora.AnaforaAnnotation):
                                text_type_attrib_map[annotation_regex][annotation.type][key][value] += 1
                                default_type_attrib_map[annotation.type][key][value] += 1

        # convert the collected counts into a model by selecting the most common
        text_predictions = {}
        for text, entity_types in text_type_map.items():

            # select the most common type for this expression
            [(entity_type, _)] = entity_types.most_common(1)

            # only save to the model if we've seen this entity frequently enough
            if min_count is None or entity_types[entity_type] >= min_count:

                # select the most common value for each attribute
                attrib = {}
                for name, values in text_type_attrib_map[text][entity_type].items():
                    [(value, _)] = values.most_common(1)
                    attrib[name] = value

                # associate the type and attributes with the regular expression
                text_predictions[text] = (entity_type, attrib)

        # convert the global collected counts into a default model by selecting the most common
        default_predictions = {}
        for entity_type in default_type_attrib_map:
            attrib = default_predictions[entity_type] = {}
            for name, values in default_type_attrib_map[entity_type].items():
                [(value, _)] = values.most_common(1)
                attrib[name] = value

        return cls(text_predictions, default_predictions)

    def __init__(self, regex_type_attributes_map, default_type_attributes_map=None):
        self.regex_type_attributes_map = regex_type_attributes_map
        self.default_type_attributes_map = default_type_attributes_map

    def __eq__(self, other):
        return (self.regex_type_attributes_map == other.regex_type_attributes_map and
                self.default_type_attributes_map == other.default_type_attributes_map)

    def __repr__(self):
        name = self.__class__.__name__
        if self.default_type_attributes_map is None:
            return '{0}({1})'.format(name, self.regex_type_attributes_map)
        else:
            return '{0}({1}, {2})'.format(name, self.regex_type_attributes_map, self.default_type_attributes_map)

    def annotate(self, text, data):
        """
        Adds annotations by matching the model's regular expressions against the text.

        :param str text: the text to be annotated
        :param anafora.AnaforaData data: the data to which the annotations should be added
        """

        # index any existing annotations so we can add to them if necessary
        span_type_annotation_map = {}
        for annotation in data.annotations:
            span_type_annotation_map[annotation.spans, annotation.type] = annotation
            if self.default_type_attributes_map is not None:
                if annotation.type in self.default_type_attributes_map:
                    for key, value in self.default_type_attributes_map[annotation.type].items():
                        if key not in annotation.properties:
                            annotation.properties[key] = value

        # create an overall regular expression where longest expressions are matched first
        # NOTE: we have to use the regex library, not the re library, because we need more that 100 groups
        patterns = sorted(self.regex_type_attributes_map, key=len, reverse=True)
        pattern = regex.compile('|'.join('({0})'.format(pattern) for pattern in patterns))

        # for each match, create an annotation with the appropriate type and attributes
        for i, match in enumerate(pattern.finditer(text)):
            pattern = patterns[match.lastindex - 1]
            entity_type, attributes = self.regex_type_attributes_map[pattern]
            spans = ((match.start(), match.end()),)
            key = (spans, entity_type)
            if key in span_type_annotation_map:
                entity = span_type_annotation_map[key]
            else:
                entity = anafora.AnaforaEntity()
                entity.id = "{0}@regex".format(i)
                entity.type = entity_type
                entity.spans = spans
                data.annotations.append(entity)
            for key, value in attributes.items():
                entity.properties[key] = value

    def prune_by_precision(self, min_precision, text_data_pairs):
        """
        Removes patterns from the model that don't reach a minimum precision

        :param float min_precision: the minimum precision required of a pattern when applied to the given data
        :param collections.Iterable text_data_pairs: an iterable of `(text, data)` pairs where `text` is a string and
            `data` is an anafora.AnaforaData object
        """
        pattern_scores = collections.defaultdict(lambda: anafora.evaluate.Scores())
        for text, data in text_data_pairs:

            # collect the spans of each type of reference annotation
            reference_type_spans_map = collections.defaultdict(lambda: set())
            for annotation in data.annotations:
                reference_type_spans_map[annotation.type].add(annotation.spans)

            # make predictions with each pattern in the model
            for pattern in self.regex_type_attributes_map:
                predicted_spans = {((m.start(), m.end()),) for m in regex.finditer(pattern, text)}
                if predicted_spans:
                    predicted_type, _ = self.regex_type_attributes_map[pattern]

                    # update the scores for this pattern
                    pattern_scores[pattern].add(reference_type_spans_map[predicted_type], predicted_spans)

        # delete any pattern with a precision lower than the minimum requested
        for pattern, scores in pattern_scores.items():
            if scores.precision() < min_precision:
                del self.regex_type_attributes_map[pattern]

    def to_file(self, path_or_file):
        """
        :param string|file path_or_file: a string path or a file object where the RegexAnnotator should be serialized
        """

        # if passed a path instead of a file, open the path as a file for writing and recursively call to_file
        try:
            write = path_or_file.write
        except AttributeError:
            with codecs.open(path_or_file, 'w', 'utf-8') as output_file:
                self.to_file(output_file)
        else:

            # write out each pattern as a line in the model file: <regex>\t<type>\t<attributes>
            for entity_type, attributes in sorted(self.default_type_attributes_map.items()):
                write('\t')
                write(entity_type)
                write('\t')
                write(json.dumps(attributes))
                write('\n')
            for expression, (entity_type, attributes) in sorted(self.regex_type_attributes_map.items()):
                write(expression)
                write('\t')
                write(entity_type)
                write('\t')
                write(json.dumps(attributes))
                write('\n')


def _train(train_dir, model_file, text_dir=None, xml_name_regex="[.]xml$", text_encoding="utf-8",
           min_count=None, min_precision=None):

    # returns an iterator over (text, data) pairs
    def text_data_pairs():
        for sub_dir, text_name, xml_names in anafora.walk(train_dir, xml_name_regex):
            if text_dir is not None:
                text_path = os.path.join(text_dir, text_name)
            else:
                text_path = os.path.join(train_dir, sub_dir, text_name)
            if not os.path.exists(text_path):
                logging.warning("no text found at %s", text_path)
                continue
            with codecs.open(text_path, 'r', text_encoding) as text_file:
                text = text_file.read()
            for xml_name in xml_names:
                data = anafora.AnaforaData.from_file(os.path.join(train_dir, sub_dir, xml_name))
                yield text, data

    # train the model, prune if requested, and write it to a file
    model = RegexAnnotator.train(text_data_pairs(), min_count)
    if min_precision is not None:
        model.prune_by_precision(min_precision, text_data_pairs())
    model.to_file(model_file)


def _annotate(model_file, text_dir, output_dir, data_dir=None, xml_name_regex="[.]xml$",
              text_encoding="utf-8", extension=".system.completed.xml"):

    if text_dir is not None:
        iterator = anafora.walk_flat_to_anafora(text_dir)
    elif data_dir is not None:
        iterator = anafora.walk_anafora_to_anafora(data_dir, xml_name_regex)
    else:
        iterator = anafora.walk_anafora_to_anafora(output_dir, xml_name_regex)

    # load a model from the file
    model = RegexAnnotator.from_file(model_file)

    # annotate each text
    for input_sub_dir, output_sub_dir, text_name, xml_names in iterator:
        if data_dir is None:
            data_iter = [(anafora.AnaforaData(), text_name + extension)]
        else:
            data_iter = [(anafora.AnaforaData.from_file(os.path.join(data_dir, input_sub_dir, xml_name)),
                          regex.sub(r'[.][^.]*[.][^.]*[.][^.]*[.]xml', extension, xml_name))
                         for xml_name in xml_names]

        for data, output_name in data_iter:
            # read in the text
            if text_dir is not None:
                text_path = os.path.join(text_dir, text_name)
            elif data_dir is not None:
                text_path = os.path.join(data_dir, input_sub_dir, text_name)
            else:
                text_path = os.path.join(output_dir, input_sub_dir, text_name)
            with codecs.open(text_path, 'r', text_encoding) as text_file:
                text = text_file.read()

            # annotate the text
            model.annotate(text, data)

            # save the annotated data to the output directory
            data_output_dir = os.path.join(output_dir, output_sub_dir)
            if not os.path.exists(data_output_dir):
                os.makedirs(data_output_dir)
            data_output_path = os.path.join(data_output_dir, output_name)
            data.indent()
            data.to_file(data_output_path)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers()

    train_parser = subparsers.add_parser("train", help="Train a regex model from a set of annotated entities")
    train_parser.set_defaults(func=_train)
    train_parser.add_argument("-i", "--input", metavar="DIR", required=True, dest="train_dir",
                              help="The root of a set of Anafora XML files containing entity annotations.")
    train_parser.add_argument("-x", "--xml-name-regex", metavar="REGEX", default="[.]xml$",
                              help="A regular expression for matching XML files in the input subdirectories " +
                                   "(default: %(default)r)")
    train_parser.add_argument("-m", "--model", metavar="FILE", dest="model_file", required=True,
                              help="The file where the trained regex model should be written.")
    train_parser.add_argument("-t", "--text", metavar="DIR", dest="text_dir",
                              help="A flat directory containing the raw text. By default, the input directory is " +
                                   "assumed to contain the raw text alongside the Anafora XMLs.")
    train_parser.add_argument("-te", "--text-encoding", metavar="ENCODING", default="utf-8",
                              help="The encoding of the text files, important for correct character offsets " +
                                   "(default: %(default)s)")
    train_parser.add_argument("-mc", "--min-count", metavar="N", type=int,
                              help="Only keep regular expressions that matched at least N entities.")
    train_parser.add_argument("-mp", "--min-precision", metavar="X", type=float,
                              help="Only keep regular expressions with precision of at least X on the training data")

    annotate_parser = subparsers.add_parser("annotate", help="Predict entity annotations in data using a regex model")
    annotate_parser.set_defaults(func=_annotate)
    annotate_parser.add_argument("-m", "--model", metavar="FILE", dest="model_file", required=True,
                                 help="The file containing the trained regex model.")
    annotate_parser.add_argument("-t", "--text", metavar="DIR", dest="text_dir",
                                 help="The raw text that should be annotated with the regex model")
    annotate_parser.add_argument("-d", "--data", metavar="DIR", dest="data_dir",
                                 help="A directory of pre-annotated Anafora XMLs for the given raw text")
    annotate_parser.add_argument("-x", "--xml-name-regex", metavar="REGEX", default="[.]xml$",
                                 help="A regular expression for matching XML files in the subdirectories, typically " +
                                      "used to restrict the evaluation to a subset of the available files (default: " +
                                      "%(default)r)")
    annotate_parser.add_argument("-o", "--output", metavar="DIR", required=True, dest="output_dir",
                                 help="The directory where the Anafora XML files containing the model predictions " +
                                      "should be written.")
    annotate_parser.add_argument("-e", "--extension", metavar="EXT", default=".system.completed.xml",
                                 help="The suffix that should be given to the model prediction files " +
                                      "(default: %(default)r)")
    annotate_parser.add_argument("-te", "--text-encoding", metavar="ENCODING", default="utf-8",
                                 help="The encoding of the text files (important for correct character offsets).")

    args = parser.parse_args()
    kwargs = vars(args)
    kwargs.pop("func")(**kwargs)



import argparse
import collections
import logging
import os

import anafora


class Schema(object):
    def __init__(self, xml):
        """
        :param xml.etree.ElementTree.Element xml: the <schema> element
        """
        default_attribute_elem = xml.find("defaultattribute")
        definition_elem = xml.find("definition")
        entities_elems = definition_elem.findall("entities")
        relations_elems = definition_elem.findall("relations")
        annotations_elems = entities_elems + relations_elems
        if not annotations_elems:
            raise ValueError("no entities or relations in schema")

        self.default_attributes = {}
        if default_attribute_elem is not None:
            for attribute_elem in default_attribute_elem:
                self.default_attributes[attribute_elem.tag] = attribute_elem.text

        self.type_to_properties = {}
        for annotations_elem in annotations_elems:
            if annotations_elem is not None:
                for annotation_elem in annotations_elem:
                    annotation_type = annotation_elem.attrib["type"]
                    properties = {}
                    properties_elem = annotation_elem.find("properties")
                    if properties_elem is not None:
                        for property_elem in properties_elem:
                            schema_property = SchemaProperty(property_elem, self.default_attributes)
                            properties[schema_property.type] = schema_property
                    self.type_to_properties[annotation_type] = properties

    @classmethod
    def from_file(cls, xml_path):
        return cls(anafora.ElementTree.parse(xml_path).getroot())

    def validate(self, annotation):
        """
        :param AnaforaAnnotation annotation: the annotation to be validated
        """
        if annotation.is_self_referential():
            msg = '{0!r} is self-referential'
            raise SchemaValidationError(msg.format(annotation.id))
        schema_properties = self.type_to_properties.get(annotation.type)
        if schema_properties is None:
            msg = 'invalid annotation type {0!r}'
            raise SchemaValidationError(msg.format(annotation.type))
        for schema_property in schema_properties.values():
            if schema_property.required and not schema_property.type in annotation.properties:
                msg = 'missing required property {0!r} of annotation type {1!r}'
                raise SchemaValidationError(msg.format(schema_property.type, annotation.type))
        for name, value in annotation.properties.items():
            if name not in schema_properties:
                msg = 'no property {0!r} defined for annotation type {1!r}'
                raise SchemaValidationError(msg.format(name, annotation.type))
            schema_property = schema_properties[name]
            if schema_property.instance_of is not None:
                if value is None:
                    if schema_property.required:
                        msg = 'missing value for property {0!r} of annotation type {1!r}'
                        raise SchemaValidationError(msg.format(schema_property.type, annotation.type))
                elif not isinstance(value, anafora.AnaforaAnnotation):
                    msg = 'invalid value {0!r} for property {1!r} of annotation type {2!r}'
                    raise SchemaValidationError(msg.format(value, schema_property.type, annotation.type))
                elif not value.type in schema_property.instance_of:
                    msg = 'invalid type {0!r} for property {1!r} of annotation type {2!r}'
                    raise SchemaValidationError(msg.format(value.type, schema_property.type, annotation.type))
            if schema_property.choices is not None:
                if isinstance(value, anafora.AnaforaAnnotation):
                    msg = 'invalid value {0!r} for property {1!r} of annotation type {2!r}'
                    raise SchemaValidationError(msg.format(value, schema_property.type, annotation.type))
                elif value not in schema_property.choices:
                    msg = 'invalid value {0!r} for property {1!r} of annotation type {2!r}'
                    raise SchemaValidationError(msg.format(value, schema_property.type, annotation.type))

    def errors(self, data):
        """
        :param AnaforaData data: the data to be validated
        :return: a list of (invalid annotation, explanation string)
        """
        errors = []
        for annotation in data.annotations:
            try:
                self.validate(annotation)
            except SchemaValidationError as error:
                errors.append((annotation, str(error)))
        return errors


class SchemaProperty(object):
    def __init__(self, xml, default_attributes):
        """
        :param xml.etree.ElementTree.Element xml: the <property> element
        :param dict default_attributes: a mapping holding default attribute (name, value) pairs
        """

        def get(name, default):
            return xml.attrib.get(name, default_attributes.get(name, default))

        self.type = get("type", None)
        self.required = get("required", None) == "True"
        self.instance_of = get("instanceOf", None)
        if self.instance_of is not None:
            self.instance_of = self.instance_of.split(",")
        if get("input", None) == "choice":
            self.choices = xml.text.split(",")
        else:
            self.choices = None


class SchemaValidationError(RuntimeError):
    pass


def log_schema_errors(schema, anafora_dir, xml_name_regex):
    """
    :param Schema schema: the schema to validate against
    :param string anafora_dir: the Anafora directory containing directories to validate
    """
    for sub_dir, text_name, xml_names in anafora.walk(anafora_dir, xml_name_regex):
        for xml_name in xml_names:
            xml_path = os.path.join(anafora_dir, sub_dir, xml_name)
            try:
                data = anafora.AnaforaData.from_file(xml_path)
            except anafora.ElementTree.ParseError:
                logging.error("%s: invalid XML", xml_path)
            except Exception as e:
                logging.error("%s: %s", xml_path, e)
            else:
                for annotation, error in schema.errors(data):
                    logging.warn("%s: %s", xml_path, error)


def find_entities_with_identical_spans(data):
    """
    :param AnaforaData data: the Anafora data to be searched
    """
    span_entities = collections.defaultdict(lambda: [])
    for ann in data.annotations:
        if isinstance(ann, anafora.AnaforaEntity):
            span_entities[ann.spans].append(ann)
    for span, annotations in span_entities.items():
        if len(annotations) > 1:
            yield span, annotations


def log_entities_with_identical_spans(anafora_dir, xml_name_regex):
    """
    :param AnaforaData data: the Anafora data to be searched
    """
    for sub_dir, text_name, xml_names in anafora.walk(anafora_dir, xml_name_regex):
        for xml_name in xml_names:
            xml_path = os.path.join(anafora_dir, sub_dir, xml_name)
            try:
                data = anafora.AnaforaData.from_file(xml_path)
            except anafora.ElementTree.ParseError:
                pass
            else:
                for span, annotations in find_entities_with_identical_spans(data):
                    logging.warn("%s: multiple entities for span %s:\n%s",
                                 xml_path, span, "\n".join(str(ann).rstrip() for ann in annotations))


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="""%(prog)s validates Anafora XML files against an Anafora schema and
        logs any errors. It can also identify other potential errors such as the presence of distinct entities with
        identical text spans.""")
    parser.add_argument("-s", "--schema", metavar="FILE", required=True,
                        help="An Anafora schema XML file against which Anafora annotation XML files should be " +
                             "validated.")
    parser.add_argument("-i", "--input", metavar="DIR", required=True, dest="anafora_dir",
                        help="The root of a set of Anafora annotation XML directories.")
    parser.add_argument("-x", "--xml-name-regex", metavar="REGEX", default="[.]xml$",
                        help="A regular expression for matching XML files, typically used to restrict the validation " +
                             "to a subset of the available files (default: %(default)r)")
    parser.add_argument("--identical-spans", action='store_true',
                        help="Also log any pairs of entities that span the exact same text offsets.")
    args = parser.parse_args()
    logging.basicConfig(format="%(levelname)s:%(message)s")

    log_schema_errors(Schema.from_file(args.schema), args.anafora_dir, args.xml_name_regex)
    if args.identical_spans:
        log_entities_with_identical_spans(args.anafora_dir, args.xml_name_regex)



import re

import anafora
import anafora.timeml

APW19981205_0374_tml = """<?xml version="1.0" ?>
<TimeML xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:noNamespaceSchemaLocation="http://timeml.org/timeMLdocs/TimeML_1.2.1.xsd">

<DOCID>APW19981205.0374</DOCID>

<DCT><TIMEX3 tid="t0" type="TIME" functionInDocument="CREATION_TIME" temporalFunction="false" value="1998-12-05T09:42:00">12/05/1998 09:42:00</TIMEX3></DCT>

<TITLE>Poland to participate in preparations for NATO summit</TITLE>


<EXTRAINFO>
APW19981205.0374 <DOCTYPE> NEWS STORY </DOCTYPE><DATE_TIME> 12/05/1998 09:42:00 </DATE_TIME><HEADER>w1008 &amp;Cx1f; wstm-
r i &amp;Cx13;  &amp;Cx11;  BC-Poland-Germany-NATO     12-05 0152</HEADER><SLUG> BC-Poland-Germany-NATO </SLUG><HEADLINE>Poland to participate in preparations for NATO summit</HEADLINE> &amp;UR; AP Photos WAR150-1 &amp;QL;
</EXTRAINFO>


<TEXT>

WARSAW, Poland (AP) _

Poland, Hungary and the Czech Republic can now <EVENT eid="e1" class="OCCURRENCE">participate</EVENT> in NATO strategy <EVENT eid="e24" class="OCCURRENCE">planning</EVENT>, because all NATO member states have <EVENT eid="e2" class="I_ACTION">approved</EVENT> their <EVENT eid="e3" class="STATE">joining</EVENT> the alliance, German defense minister Rudolf Scharping <EVENT eid="e4" class="REPORTING">said</EVENT> <TIMEX3 tid="t2" type="DATE" functionInDocument="NONE" temporalFunction="true" value="1998-12-05">Saturday</TIMEX3>.

The Netherlands <TIMEX3 tid="t3" type="DATE" functionInDocument="NONE" temporalFunction="true" value="1998-W49" anchorTimeID="t0">early this week</TIMEX3> <EVENT eid="e5" class="STATE">became</EVENT> the last country to <EVENT eid="e7" class="OCCURRENCE">ratify</EVENT> NATO expansion to include the three former Warsaw Pact nations.

The three are <EVENT eid="e8" class="I_STATE">expected</EVENT> to <EVENT eid="e10" class="OCCURRENCE">join</EVENT> <SIGNAL sid="s3">in</SIGNAL> <TIMEX3 tid="t4" type="DATE" functionInDocument="NONE" temporalFunction="true" value="1999-04">April</TIMEX3>, as part of <EVENT eid="e11" class="OCCURRENCE">marking</EVENT> NATO's 50th anniversary. With their membership secure, they also can <EVENT eid="e12" class="OCCURRENCE">participate</EVENT> in <EVENT eid="e13" class="I_ACTION">planning</EVENT> for the NATO <EVENT eid="e14" class="OCCURRENCE">summit</EVENT> <SIGNAL sid="s4">in</SIGNAL> <TIMEX3 tid="t5" type="DATE" functionInDocument="NONE" temporalFunction="true" value="1999-04">April</TIMEX3>.

Scharping, <EVENT eid="e15" class="ASPECTUAL">ending</EVENT> a <TIMEX3 tid="t6" type="DURATION" functionInDocument="NONE" temporalFunction="true" value="P2D">two-day</TIMEX3>  <EVENT eid="e16" class="OCCURRENCE">visit</EVENT>, <EVENT eid="e17" class="REPORTING">said</EVENT> Poland was well <EVENT eid="e18" class="I_STATE">prepared</EVENT> to <EVENT eid="e19" class="OCCURRENCE">join</EVENT> the alliance. Germany has <EVENT eid="e21" class="STATE">been</EVENT> a strong advocate of Poland's access to NATO, <EVENT eid="e22" class="REPORTING">saying</EVENT> it will <EVENT eid="e23" class="OCCURRENCE">serve</EVENT> European security and stability.

</TEXT>


<EXTRAINFO>
&amp;UR; (ms-aet)  (PROFILE (WS SL:BC-Poland-Germany-NATO; CT:i; (REG:EURO;) (REG:BRIT;) (REG:SCAN;) (REG:ENGL;) (LANG:ENGLISH;)) )
<TRAILER> AP-NY-12-05-98 0942EST </TRAILER>
</EXTRAINFO>


<MAKEINSTANCE eiid="ei1" eventID="e1" tense="PRESENT" aspect="NONE" pos="UNKNOWN" modality="can" polarity="POS"/>
<MAKEINSTANCE eiid="ei2" eventID="e2" tense="PAST" aspect="PERFECTIVE" pos="UNKNOWN" polarity="POS"/>
<MAKEINSTANCE eiid="ei3" eventID="e3" tense="NONE" aspect="NONE" pos="NOUN" polarity="POS"/>
<MAKEINSTANCE eiid="ei4" eventID="e4" tense="PAST" aspect="NONE" pos="UNKNOWN" polarity="POS"/>
<MAKEINSTANCE eiid="ei5" eventID="e5" tense="PAST" aspect="NONE" pos="UNKNOWN" polarity="POS"/>
<MAKEINSTANCE eiid="ei7" eventID="e7" tense="INFINITIVE" aspect="NONE" pos="VERB" polarity="POS"/>
<MAKEINSTANCE eiid="ei9" eventID="e8" tense="PRESENT" aspect="NONE" pos="UNKNOWN" polarity="POS"/>
<MAKEINSTANCE eiid="ei12" eventID="e10" tense="INFINITIVE" aspect="NONE" pos="VERB" polarity="POS"/>
<MAKEINSTANCE eiid="ei13" eventID="e11" tense="NONE" aspect="NONE" pos="NOUN" polarity="POS"/>
<MAKEINSTANCE eiid="ei14" eventID="e12" tense="PRESENT" aspect="NONE" pos="UNKNOWN" modality="can" polarity="POS"/>
<MAKEINSTANCE eiid="ei15" eventID="e13" tense="NONE" aspect="NONE" pos="NOUN" polarity="POS"/>
<MAKEINSTANCE eiid="ei16" eventID="e14" tense="NONE" aspect="NONE" pos="NOUN" polarity="POS"/>
<MAKEINSTANCE eiid="ei17" eventID="e15" tense="PRESPART" aspect="NONE" pos="VERB" polarity="POS"/>
<MAKEINSTANCE eiid="ei18" eventID="e16" tense="NONE" aspect="NONE" pos="NOUN" polarity="POS"/>
<MAKEINSTANCE eiid="ei19" eventID="e17" tense="PAST" aspect="NONE" pos="UNKNOWN" polarity="POS"/>
<MAKEINSTANCE eiid="ei20" eventID="e18" tense="PAST" aspect="NONE" pos="UNKNOWN" polarity="POS"/>
<MAKEINSTANCE eiid="ei21" eventID="e19" tense="INFINITIVE" aspect="NONE" pos="VERB" polarity="POS"/>
<MAKEINSTANCE eiid="ei23" eventID="e21" tense="PRESENT" aspect="PERFECTIVE" pos="UNKNOWN" polarity="POS"/>
<MAKEINSTANCE eiid="ei24" eventID="e22" tense="PRESPART" aspect="NONE" pos="VERB" polarity="POS"/>
<MAKEINSTANCE eiid="ei25" eventID="e23" tense="FUTURE" aspect="NONE" pos="UNKNOWN" polarity="POS"/>
<MAKEINSTANCE eiid="ei26" eventID="e24" tense="NONE" aspect="NONE" pos="NOUN" polarity="POS"/>
<TLINK lid="l2" timeID="t0" relatedToEventInstance="ei1" relType="INCLUDES" origin="USER"/>
<TLINK lid="l3" eventInstanceID="ei2" relatedToEventInstance="ei1" relType="BEFORE" origin="USER"/>
<TLINK lid="l5" timeID="t0" relatedToEventInstance="ei2" relType="AFTER" origin="USER"/>
<TLINK lid="l6" timeID="t2" relatedToEventInstance="ei4" relType="INCLUDES" origin="USER"/>
<TLINK lid="l7" timeID="t2" relatedToEventInstance="ei4" relType="INCLUDES" origin="USER"/>
<TLINK lid="l8" eventInstanceID="ei4" relatedToEventInstance="ei2" relType="AFTER" origin="USER"/>
<TLINK lid="l9" timeID="t3" relatedToEventInstance="ei5" relType="INCLUDES" origin="USER"/>
<TLINK lid="l11" eventInstanceID="ei5" relatedToEventInstance="ei7" relType="SIMULTANEOUS" origin="USER"/>
<TLINK lid="l12" timeID="t0" relatedToEventInstance="ei9" relType="INCLUDES" origin="USER"/>
<TLINK lid="l14" timeID="t4" signalID="s3" relatedToEventInstance="ei12" relType="INCLUDES" origin="USER"/>
<TLINK lid="l15" timeID="t4" relatedToEventInstance="ei13" relType="INCLUDES" origin="USER"/>
<TLINK lid="l16" timeID="t0" relatedToEventInstance="ei13" relType="BEFORE" origin="USER"/>
<TLINK lid="l17" timeID="t0" relatedToEventInstance="ei14" relType="INCLUDES" origin="USER"/>
<TLINK lid="l20" timeID="t5" signalID="s4" relatedToEventInstance="ei16" relType="INCLUDES" origin="USER"/>
<TLINK lid="l21" timeID="t6" relatedToEventInstance="ei18" relType="INCLUDES" origin="USER"/>
<TLINK lid="l23" timeID="t2" relatedToEventInstance="ei19" relType="INCLUDES" origin="USER"/>
<TLINK lid="l24" timeID="t6" relatedToEventInstance="ei19" relType="DURING" origin="USER"/>
<TLINK lid="l26" eventInstanceID="ei3" relatedToTime="t4" relType="INCLUDES" origin="USER"/>
<TLINK lid="l27" timeID="t0" relatedToEventInstance="ei26" relType="BEFORE" origin="USER"/>
<TLINK lid="l28" timeID="t0" relatedToEventInstance="ei15" relType="BEFORE" origin="USER"/>
<TLINK lid="l29" timeID="t2" relatedToEventInstance="ei17" relType="AFTER" origin="USER"/>
<TLINK lid="l32" timeID="t4" relatedToEventInstance="ei21" relType="INCLUDES" origin="USER"/>
<TLINK lid="l33" timeID="t0" relatedToEventInstance="ei20" relType="AFTER" origin="USER"/>
<TLINK lid="l35" timeID="t0" relatedToEventInstance="ei23" relType="AFTER" origin="USER"/>
<TLINK lid="l36" timeID="t0" relatedToEventInstance="ei24" relType="AFTER" origin="USER"/>
<TLINK lid="l38" timeID="t0" relatedToEventInstance="ei25" relType="BEFORE" origin="USER"/>
<SLINK lid="l1" eventInstanceID="ei2" subordinatedEventInstance="ei3" relType="MODAL"/>
<SLINK lid="l10" eventInstanceID="ei1" subordinatedEventInstance="ei26" relType="MODAL"/>
<SLINK lid="l13" eventInstanceID="ei9" subordinatedEventInstance="ei12" relType="MODAL"/>
<SLINK lid="l18" eventInstanceID="ei14" subordinatedEventInstance="ei15" relType="MODAL"/>
<SLINK lid="l19" eventInstanceID="ei15" subordinatedEventInstance="ei16" relType="MODAL"/>
<SLINK lid="l30" eventInstanceID="ei19" subordinatedEventInstance="ei20" relType="EVIDENTIAL"/>
<SLINK lid="l31" eventInstanceID="ei20" subordinatedEventInstance="ei21" relType="MODAL"/>
<SLINK lid="l37" eventInstanceID="ei24" subordinatedEventInstance="ei25" relType="EVIDENTIAL"/>
<ALINK lid="l22" eventInstanceID="ei18" relatedToEventInstance="ei17" relType="CULMINATES"/>
</TimeML>
"""


def test_to_text(tmpdir):

    # remove elements
    text = re.sub(r"<[^>]*>", "", APW19981205_0374_tml)
    # replace XML entities
    text = text.replace("&amp;", "&")
    # remove spaces before and after root element
    text = text[1:-1]

    # test that method produces the same thing
    path = tmpdir.join("APW19981205_0374.tml")
    path.write(APW19981205_0374_tml)
    assert anafora.timeml.to_text(str(path)) == text


def test_to_document_creation_time(tmpdir):
    path = tmpdir.join("APW19981205_0374.tml")
    path.write(APW19981205_0374_tml)
    assert anafora.timeml.to_document_creation_time(str(path)) == "1998-12-05T09:42:00"



def test_to_anafora_data(tmpdir):
    path = tmpdir.join("APW19981205_0374.tml")
    path.write(APW19981205_0374_tml)

    text = anafora.timeml.to_text(str(path))
    data = anafora.timeml.to_anafora_data(str(path))

    # check counts of various annotations
    assert len(list(data.annotations.select_type("EVENT"))) == 21
    assert len(list(data.annotations.select_type("TIMEX3"))) == 6
    assert len(list(data.annotations.select_type("SIGNAL"))) == 2
    assert len(list(data.annotations.select_type("TLINK"))) == 26
    assert len(list(data.annotations.select_type("SLINK"))) == 8
    assert len(list(data.annotations.select_type("ALINK"))) == 1

    # <TIMEX3 tid="t0" type="TIME" functionInDocument="CREATION_TIME" temporalFunction="false"
    #         value="1998-12-05T09:42:00">12/05/1998 09:42:00</TIMEX3>
    annotation = data.annotations.select_id("1@e@APW19981205_0374@gold")
    for start, end in annotation.spans:
        assert text[start:end] == "12/05/1998 09:42:00"
    pattern = "^<entity><id>.*?</id><type>TIMEX3</type><span>20,39</span><properties>.*?</properties></entity>$"
    assert re.match(pattern, str(annotation))
    assert dict(annotation.properties.items()) == {
        "type": "TIME",
        "functionInDocument": "CREATION_TIME",
        "temporalFunction": "false",
        "value": "1998-12-05T09:42:00",
    }

    # <TIMEX3 tid="t3" type="DATE" functionInDocument="NONE" temporalFunction="true" value="1998-W49"
    #         anchorTimeID="t0">early this week</TIMEX3>
    annotation = data.annotations.select_id("8@e@APW19981205_0374@gold")
    for start, end in annotation.spans:
        assert text[start:end] == "early this week"
    pattern = "^<entity><id>.*?</id><type>TIMEX3</type><span>.*?</span><properties>.*?</properties></entity>$"
    assert re.match(pattern, str(annotation))
    assert dict(annotation.properties.items()) == {
        "type": "DATE",
        "functionInDocument": "NONE",
        "temporalFunction": "true",
        "value": "1998-W49",
        "anchorTimeID": data.annotations.select_id("1@e@APW19981205_0374@gold"),
    }

    # <EVENT eid="e8" class="I_STATE">expected</EVENT>
    annotation = data.annotations.select_id("11@e@APW19981205_0374@gold")
    for start, end in annotation.spans:
        assert text[start:end] == "expected"
    pattern = "^<entity><id>.*?</id><type>EVENT</type><span>.*?</span><properties>.*?</properties></entity>$"
    assert re.match(pattern, str(annotation))
    assert dict(annotation.properties.items()) == {"class": "I_STATE"}

    # <SIGNAL sid="s3">in</SIGNAL>
    annotation = data.annotations.select_id("13@e@APW19981205_0374@gold")
    for start, end in annotation.spans:
        assert text[start:end] == "in"
    pattern = "^<entity><id>.*?</id><type>SIGNAL</type><span>.*?</span></entity>$"
    assert re.match(pattern, str(annotation))
    assert not annotation.properties.items()

    # <MAKEINSTANCE eiid="ei23" eventID="e21" tense="PRESENT" aspect="PERFECTIVE" pos="UNKNOWN" polarity="POS"/>
    annotation = data.annotations.select_id("47@r@APW19981205_0374@gold")
    pattern = "^<relation><id>.*?</id><type>MAKEINSTANCE</type><properties>.*?</properties></relation>"
    assert re.match(pattern, str(annotation))
    assert dict(annotation.properties.items()) == {
        "eventID": data.annotations.select_id("27@e@APW19981205_0374@gold"),
        "tense": "PRESENT",
        "aspect": "PERFECTIVE",
        "pos": "UNKNOWN",
        "polarity": "POS",
    }
    assert dict(annotation.properties["eventID"].properties.items()) == {"class": "STATE"}

    # <TLINK lid="l9" timeID="t3" relatedToEventInstance="ei5" relType="INCLUDES" origin="USER"/>
    annotation = data.annotations.select_id("57@r@APW19981205_0374@gold")
    pattern = "^<relation><id>.*?</id><type>TLINK</type><properties>.*?</properties></relation>"
    assert re.match(pattern, str(annotation))
    assert dict(annotation.properties.items()) == {
        "timeID": data.annotations.select_id("8@e@APW19981205_0374@gold"),
        "relatedToEventInstance": data.annotations.select_id("34@r@APW19981205_0374@gold"),
        "relType": "INCLUDES",
        "origin": "USER",
    }
    assert dict(annotation.properties["timeID"].properties.items()) == {
        "type": "DATE",
        "functionInDocument": "NONE",
        "temporalFunction": "true",
        "value": "1998-W49",
        "anchorTimeID": data.annotations.select_id("1@e@APW19981205_0374@gold"),
    }
    assert dict(annotation.properties["relatedToEventInstance"].properties.items()) == {
        "eventID": data.annotations.select_id("9@e@APW19981205_0374@gold"),
        "tense": "PAST",
        "aspect": "NONE",
        "pos": "UNKNOWN",
        "polarity": "POS",
    }

    # <SLINK lid="l37" eventInstanceID="ei24" subordinatedEventInstance="ei25" relType="EVIDENTIAL"/>
    annotation = data.annotations.select_id("84@r@APW19981205_0374@gold")
    pattern = "^<relation><id>.*?</id><type>SLINK</type><properties>.*?</properties></relation>"
    assert re.match(pattern, str(annotation))
    assert dict(annotation.properties.items()) == {
        "eventInstanceID": data.annotations.select_id("48@r@APW19981205_0374@gold"),
        "subordinatedEventInstance": data.annotations.select_id("49@r@APW19981205_0374@gold"),
        "relType": "EVIDENTIAL",
    }

    # <ALINK lid="l22" eventInstanceID="ei18" relatedToEventInstance="ei17" relType="CULMINATES"/>
    annotation = data.annotations.select_id("85@r@APW19981205_0374@gold")
    pattern = "^<relation><id>.*?</id><type>ALINK</type><properties>.*?</properties></relation>"
    assert re.match(pattern, str(annotation))
    assert dict(annotation.properties.items()) == {
        "eventInstanceID": data.annotations.select_id("43@r@APW19981205_0374@gold"),
        "relatedToEventInstance": data.annotations.select_id("42@r@APW19981205_0374@gold"),
        "relType": "CULMINATES",
    }
